package com.terraformersmc.terrestria.feature.trees;

import com.mojang.datafixers.Dynamic;
import com.terraformersmc.terrestria.feature.TreeDefinition;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2944;
import net.minecraft.class_3111;
import net.minecraft.class_3341;
import net.minecraft.class_3532;
import net.minecraft.class_3746;
import net.minecraft.class_3747;

public class RubberTreeFeature extends class_2944<class_3111> {
	private TreeDefinition.Basic tree;

	public RubberTreeFeature(Function<Dynamic<?>, ? extends class_3111> function, boolean notify, TreeDefinition.Basic tree) {
		super(function, notify);

		this.tree = tree;
	}

	public RubberTreeFeature sapling() {
		return new RubberTreeFeature(class_3111::method_13565, true, tree);
	}

	@Override
	public boolean method_12775(Set<class_2338> blocks, class_3747 world, Random rand, class_2338 origin, class_3341 boundingBox) {
		// Total trunk height
		int height = rand.nextInt(4) + 12;

		if (origin.method_10264() + height + 1 > 256 || origin.method_10264() < 1) {
			return false;
		}

		class_2338 below = origin.method_10074();

		if (!method_16430(world, below)) {
			return false;
		}

		if (!checkForObstructions(world, origin, height)) {
			return false;
		}

		method_16427(world, below);
		growTrunk(blocks, world, new class_2338.class_2339(origin), height, boundingBox);
		growBranches(blocks, world, new class_2338.class_2339(origin), height, rand, boundingBox);

		return true;
	}

	private boolean checkForObstructions(class_3746 world, class_2338 origin, int height) {
		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for (int dY = 0; dY < height; dY++) {
			pos.method_10099(origin.method_10264() + dY);

			if (!method_16432(world, pos)) {
				return false;
			}
		}

		return true;
	}

	private void growTrunk(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, class_3341 boundingBox) {
		int x = pos.method_10263();
		int z = pos.method_10260();

		for (int i = 0; i < height; i++) {
			pos.method_10103(x, pos.method_10264(), z);
			method_12773(blocks, world, pos, tree.getLog(), boundingBox);
			pos.method_10098(class_2350.field_11036);
		}
	}

	private void growBranches(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, Random random, class_3341 boundingBox) {
		int x = pos.method_10263();
		int y = pos.method_10264();
		int z = pos.method_10260();

		for (int branch = 0; branch < 16; branch++) {
			int baseY = random.nextInt(height - 2) + 4;

			float length = random.nextFloat() * 7 + 2;
			float angle = random.nextFloat() * (float) Math.PI * 2;

			int offsetX = (int) (class_3532.method_15362(angle) * length);
			int offsetZ = (int) (class_3532.method_15374(angle) * length);

			int moveX = offsetX > 0 ? 1 : -1;
			int moveZ = offsetZ > 0 ? 1 : -1;

			int absX = Math.abs(offsetX);
			int absZ = Math.abs(offsetZ);

			int movedX = 0;
			int movedZ = 0;

			float stepX = 1.0F / absX;
			float stepZ = 1.0F / absZ;

			for (int movement = 0; movement < absX + absZ; movement++) {
				if (Math.abs(movedX * stepX) < Math.abs(movedZ * stepZ) && Math.abs(movedX) < absX) {
					movedX += moveX;
				} else {
					movedZ += moveZ;
				}

				int offsetY = (int) (Math.sqrt(movedX * movedX + movedZ * movedZ) * 0.4);

				if (movedX > 2 || movedX < -2 || movedZ > 2 || movedZ < -2) {
					continue;
				}

				pos.method_10103(x + movedX, y + baseY + offsetY, z + movedZ);

				if (!method_16432(world, pos)) {
					break;
				}

				method_12773(blocks, world, pos, tree.getLog(), boundingBox);

				for (class_2350 direction : class_2350.values()) {
					pos.method_10103(x + movedX, y + baseY + offsetY, z + movedZ);
					pos.method_10098(direction);

					if (class_2944.method_16420(world, pos)) {
						method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
					}
				}
			}
		}
	}
}
