package com.terraformersmc.terrestria.feature.trees;

import com.mojang.datafixers.Dynamic;
import com.terraformersmc.terrestria.feature.TreeDefinition;
import com.terraformersmc.terraform.util.*;

import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2410;
import net.minecraft.class_2944;
import net.minecraft.class_3111;
import net.minecraft.class_3341;
import net.minecraft.class_3746;
import net.minecraft.class_3747;

public class WillowTreeFeature extends class_2944<class_3111> {
	private TreeDefinition.Basic tree;

	public WillowTreeFeature(Function<Dynamic<?>, ? extends class_3111> function, boolean notify, TreeDefinition.Basic tree) {
		super(function, notify);
		this.tree = tree;
	}

	public WillowTreeFeature sapling() {
		return new WillowTreeFeature(class_3111::method_13565, true, tree);
	}

	@Override
	public boolean method_12775(Set<class_2338> blocks, class_3747 world, Random rand, class_2338 origin, class_3341 boundingBox) {
		// Total tree height
		int height = rand.nextInt(3) + 8;

		// Maximum leaf radius.
		double maxRadius = 5 + 3 * rand.nextDouble();

		// Minimum leaf radius
		double minRadius = 1.25 + 2 * rand.nextDouble();

		// If the tree can pass the max build height
		if (origin.method_10264() + height + 1 > 256 || origin.method_10264() < 1) {
			return false;
		}

		// Make sure the tree can grow where the origin is
		class_2338 below = origin.method_10074();

		if (!method_16430(world, below)) {
			return false;
		}
		if (!checkForObstructions(world, origin, height, (int) Math.ceil(maxRadius))) {
			return false;
		}

		method_16427(world, below);

		// Grow the trunk
		growTrunk(blocks, world, new class_2338.class_2339(origin), height, boundingBox);
		growBranches(blocks, world, new class_2338.class_2339(origin.method_10079(class_2350.field_11036, height / 3)), (int) maxRadius - 1, boundingBox);
		growBranches(blocks, world, new class_2338.class_2339(origin.method_10079(class_2350.field_11036, (height / 2) + 1)), (int) (maxRadius * radiusFactor(height / 2, height)), boundingBox);

		// Grow the leaves
		class_2338.class_2339 pos = new class_2338.class_2339(origin);
		growLeaves(blocks, world, pos, height, maxRadius, minRadius, boundingBox, rand);

		return true;
	}

	// Grows the center trunk.
	private void growTrunk(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, class_3341 boundingBox) {
		// Grows the trunk at 80% of it's total height (so the trunk doesn't poke out)
		for (int i = 0; i < (height * .8); i++) {
			method_12773(blocks, world, pos, tree.getLog(), boundingBox);
			pos.method_10098(class_2350.field_11036);
		}

		// Make sure there are leaf blocks on the top of the tree
		Shapes.circle(pos, 1.5, position -> {
			if (class_2944.method_16420(world, pos)) {
				method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
			}
		});
	}

	private void growBranches(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int maxRadius, class_3341 boundingBox) {
		// Save original origin for use in z branch

		class_2338 origin = pos.method_10062();
		int branchLength = maxRadius - 2;

		pos.method_10104(class_2350.field_11039, branchLength);
		for (int x = -branchLength; x <= branchLength; x++) {
			if (method_16420(world, pos)) {
				method_12773(blocks, world, pos, tree.getLog().method_11657(class_2410.field_11459, class_2350.field_11039.method_10166()), boundingBox);
			}
			pos.method_10098(class_2350.field_11034);
		}

		pos.method_10101(origin).method_10104(class_2350.field_11043, branchLength);
		for (int z = -branchLength; z <= branchLength; z++) {
			if (method_16420(world, pos)) {
				method_12773(blocks, world, pos, tree.getLog().method_11657(class_2410.field_11459, class_2350.field_11043.method_10166()), boundingBox);
			}
			pos.method_10098(class_2350.field_11035);
		}

		// Reset the branches
		pos.method_10101(origin);
	}

	private void growDangingBit(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, class_3341 boundingBox, Random random) {
		int randHolder;

		// 33% chance of a dangling bit generating.
		if (random.nextInt(3) == 1) {
			// Length of the bit

			randHolder = random.nextInt(3);
			for (int d = 0; d < randHolder; d++) {
				pos.method_10098(class_2350.field_11033);
				method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
			}
			pos.method_10104(class_2350.field_11036, randHolder);
		}
	}

	private void growLeaves(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, double maxRadius, double minRadius, class_3341 boundingBox, Random rand) {
		int x = pos.method_10263();
		int y = pos.method_10264();
		int z = pos.method_10260();
		double radius;
		double innerRadius;

		for (int dy = 0; dy < height; dy++) {

			// If the leaves aren't 1/3 way up the trunk
			if (dy < (height / 3)) {
				continue;
			}

			pos.method_10103(x, y + dy, z);

			radius = maxRadius * radiusFactor(dy, height);
			innerRadius = minRadius * radiusFactor(dy, height);

			if (radius < 0) {
				continue;
			}

			int finalDy = dy;
			Shapes.canopyCircle(pos, radius, innerRadius, position -> {
				if (class_2944.method_16420(world, pos)) {
					method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
					if (finalDy == (height / 3)) {
						growDangingBit(blocks, world, pos, boundingBox, rand);
					}
				}
			});
		}
	}

	// Provides the factor to the radius, where x is a double from 0.0 to the height that represents the progress along the trunk.
	private double radiusFactor(double x, double height) {
		// Makes the polynomial apply to values from 0-the height
		x = x / height;

		// A 3rd-degree polynomial approximating the shape of a willow tree. from 0-1
		return 1.88 * (x * x * x) - 6.52 * (x * x) + 4.63 * x;
	}

	private boolean checkForObstructions(class_3746 world, class_2338 origin, int height, int radius) {
		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for (int dY = origin.method_10264(); dY < height; dY++) {
			for (int dZ = -radius; dZ <= radius; dZ++) {
				for (int dX = -radius; dX <= radius; dX++) {
					pos.method_10103(origin.method_10263() + dX, origin.method_10264() + dY, origin.method_10260() + dZ);

					if (!method_16432(world, pos)) {
						return false;
					}
				}
			}
		}

		pos.method_10103(origin.method_10263(), origin.method_10264() + height, origin.method_10260());

		for (int i = 0; i < 4; i++) {
			if (!method_16432(world, pos.method_10098(class_2350.field_11036))) {
				return false;
			}
		}

		return true;
	}
}
