package com.terraformersmc.terrestria.feature.trees.templates;

import com.mojang.datafixers.Dynamic;
import com.terraformersmc.terraform.block.ExtendedLeavesBlock;
import com.terraformersmc.terraform.block.QuarterLogBlock;
import com.terraformersmc.terrestria.feature.TreeDefinition;
import com.terraformersmc.terrestria.feature.trees.components.Roots;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import net.minecraft.class_2338;
import net.minecraft.class_2944;
import net.minecraft.class_3111;
import net.minecraft.class_3341;
import net.minecraft.class_3746;
import net.minecraft.class_3747;

public class ConiferTreeFeatureMega extends class_2944<class_3111> implements Roots {
	private static final int EXTRA_LEAVES_HEIGHT = 2;
	private TreeDefinition.Mega tree;
	private int height;
	private int bareTrunkHeight;

	public ConiferTreeFeatureMega(Function<Dynamic<?>, ? extends class_3111> function, boolean notify, TreeDefinition.Mega tree) {
		super(function, notify);

		this.tree = tree;
	}

	public ConiferTreeFeatureMega sapling() {
		return new ConiferTreeFeatureMega(class_3111::method_13565, true, tree);
	}

	@Override
	public boolean method_12775(Set<class_2338> blocks, class_3747 world, Random rand, class_2338 origin, class_3341 boundingBox) {
		// Total trunk height
		int height = getHeight(rand);

		// How much "bare trunk" there will be.
		int bareTrunkHeight = getBareTrunkHeight(rand);

		// Maximum leaf radius.
		// Note: Old EBXL had a maximum radius of 10, but unfortunately that would cause cascading world generation.
		// Hey, the trees are pretty massive already.
		int maxRadius = 2 + rand.nextInt(6);

		if (origin.method_10264() + height + 1 + EXTRA_LEAVES_HEIGHT > 256 || origin.method_10264() < 1) {
			return false;
		}

		for (int dZ = 0; dZ < 2; dZ++) {
			for (int dX = 0; dX < 2; dX++) {
				class_2338 below = origin.method_10069(dX, -1, dZ);

				if (!method_16430(world, below)) {
					return false;
				}
			}
		}

		if (!checkForObstructions(world, origin, height, bareTrunkHeight, maxRadius)) {
			return false;
		}

		for (int dZ = 0; dZ < 2; dZ++) {
			for (int dX = 0; dX < 2; dX++) {
				method_16427(world, origin.method_10069(dX, -1, dZ));
			}
		}

		growLeaves(blocks, world, origin, height, bareTrunkHeight, maxRadius, boundingBox);
		growTrunk(blocks, world, new class_2338.class_2339(origin), height, boundingBox);
		growRoots(blocks, world, new class_2338.class_2339(origin), bareTrunkHeight + 2, rand, boundingBox);

		return true;
	}

	private boolean checkForObstructions(class_3746 world, class_2338 origin, int height, int bareTrunkHeight, int radius) {
		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for (int i = 0; i < bareTrunkHeight; i++) {
			boolean canReplaceAll =
					method_16432(world, pos.method_10103(origin.method_10263(), origin.method_10264() + i, origin.method_10260())) &&
							method_16432(world, pos.method_10103(origin.method_10263() + 1, origin.method_10264() + i, origin.method_10260())) &&
							method_16432(world, pos.method_10103(origin.method_10263(), origin.method_10264() + i, origin.method_10260() + 1)) &&
							method_16432(world, pos.method_10103(origin.method_10263() + 1, origin.method_10264() + i, origin.method_10260() + 1));

			if (!canReplaceAll) {
				return false;
			}
		}

		for (int dY = bareTrunkHeight; dY < height + EXTRA_LEAVES_HEIGHT; dY++) {
			for (int dZ = -radius; dZ <= radius + 1; dZ++) {
				for (int dX = -radius; dX <= radius + 1; dX++) {
					pos.method_10103(origin.method_10263() + dX, origin.method_10264() + dY, origin.method_10260() + dZ);

					if (!method_16432(world, pos)) {
						return false;
					}
				}
			}
		}

		return true;
	}

	private void growLeaves(Set<class_2338> blocks, class_3747 world, class_2338 origin, int height, int bareTrunkHeight, int maxRadius, class_3341 boundingBox) {
		int radius = 0;
		int radiusTarget = 1;
		boolean topCone = true;

		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for (int dY = height + EXTRA_LEAVES_HEIGHT; dY >= bareTrunkHeight; dY--) {
			for (int dZ = -radius; dZ <= radius + 1; dZ++) {
				for (int dX = -radius; dX <= radius + 1; dX++) {
					int aZ = dZ < 1 ? -dZ : dZ - 1;
					int aX = dX < 1 ? -dX : dX - 1;

					if (radius > 0 && aZ == radius && aX == radius) {
						// Cull corners
						continue;
					}

					pos.method_10103(origin.method_10263() + dX, origin.method_10264() + dY, origin.method_10260() + dZ);

					if (class_2944.method_16420(world, pos)) {
						int distance = aZ + aX;
						int extra = dY - height + 1;

						if(extra > 0) {
							distance += extra;
						}

						method_12773(blocks, world, pos, tree.getLeaves().method_11657(ExtendedLeavesBlock.DISTANCE, Math.max(distance, 1)), boundingBox);
					}
				}
			}

			radius += 1;

			if (radius > radiusTarget) {
				if (topCone) {
					radius = 0;
					radiusTarget = Math.min(2, maxRadius);
					topCone = false;
				} else {
					radius = 1;
					radiusTarget = Math.min(radiusTarget + 1, maxRadius);
				}
			}
		}
	}

	private void growTrunk(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, class_3341 boundingBox) {
		int x = pos.method_10263();
		int y = pos.method_10264();
		int z = pos.method_10260();

		for (int i = 0; i < height; i++) {
			pos.method_10103(x, y + i, z);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.NORTHWEST), boundingBox);

			pos.method_10103(x + 1, y + i, z);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.NORTHEAST), boundingBox);

			pos.method_10103(x, y + i, z + 1);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.SOUTHWEST), boundingBox);

			pos.method_10103(x + 1, y + i, z + 1);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.SOUTHEAST), boundingBox);
		}
	}

	public int getHeight(Random rand) {
		return rand.nextInt(16) + 32;
	}

	public int getBareTrunkHeight(Random rand) {
		return 1 + rand.nextInt(12);
	}

	@Override
	public void growRoots(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int baseTrunkHeight, Random rand, class_3341 boundingBox) {

	}

	@Override
	public void tryGrowRoot(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 bottom, int baseTrunkHeight, Random rand, class_3341 boundingBox) {

	}
}
